/*
 * Copyright (c) 2004, Stefan Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 *     * Redistributions of source code must retain the above 
 *       copyright notice, this list of conditions and the 
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the 
 *       above copyright notice, this list of conditions and 
 *       the following disclaimer in the documentation and/or 
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be 
 *       used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGE.
 * 
 *
 * CONTRIBUTORS
 *  Stef Walter <stef@memberwebs.com>
 *
 */ 
 
package com.memberwebs.httpauth;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

/**
 * A pool of connections to the HttpAuth daemon.
 *
 * @author Stef Walter
 */
public class HttpAuthConnectionPool 
    implements HttpAuthConnectionSource
{
    private Stack m_connections;
    private short m_checkedout;
    private short m_max;
    
    private String m_hostname;
    private int m_port;
    private Map m_options;
    
    /**
     * Creates a new pool of connections to the HttpAuth daemon at the given
     * address. The connections will be setup to use the given handler.
     * 
     * @param hostname The address at which the HttpAuth daemon is running.
     * @param port The TCP port at which the HttpAuth daemon is listening on.
     * @param handler The HttpAuth handler to use.
     */
    public HttpAuthConnectionPool(String hostname, int port, String handler)
    {
        m_hostname = hostname;
        m_port = port;
        m_checkedout = 0;
        m_max = 16;
        m_connections = new Stack();
        m_options = new HashMap();
        m_options.put(HttpAuthConnection.OPT_HANDLER, handler);
    }
    
    /**
     * Creates a new pool of connections to the HttpAuth daemon at the given
     * address. The connections will be setup with the various options.
     * 
     * @param hostname The address at which the HttpAuth daemon is running.
     * @param port The TCP port at which the HttpAuth daemon is listening on.
     * @param options The HttpAuth options to use. This must include {@link #HttpAuthConnection.OPT_HANDLER}.
     */
    public HttpAuthConnectionPool(String hostname, int port, Map options)
    {
        m_hostname = hostname;
        m_port = port;
        m_checkedout = 0;
        m_max = 16;
        m_connections = new Stack();
        m_options = options;
    }
    
    /**
     * Set the maximum number of connections available in this pool.
     * @param max The maximum number of connections.
     */
    public void setMax(short max)
    {
        m_max = max;
    }
    
    /* ---------------------------------------------------------------------------
     * HttpAuthConnectionSource
     */
    
    /**
     * Retrieves a connection from the pool.
     * @return The connection.
     */
    public HttpAuthConnection getOpenConnection() 
        throws HttpAuthException
    {
        synchronized(this)
        {
            HttpAuthConnection conn = null;
            
            while(!m_connections.empty())
            {
                conn = (HttpAuthConnection)m_connections.pop();
                if(conn != null && conn.isConnected())
                {
                    m_checkedout++;
                    return conn;
                }
            }
            
            if(m_checkedout >= m_max)
                throw new HttpAuthException("too many connections to httpauthd: " + m_checkedout);
            
            try
            {
                // Otherwise we make a new connection
                conn = new HttpAuthConnection();
                conn.open(m_hostname, m_port, m_options);
                m_checkedout++;
                return conn; 
            }
            catch(IOException e)
            {
                throw new HttpAuthException("couldn't connect to httpauthd: " + e.getMessage());			
            }
        }
    }

    /**
     * Puts a connection back in the pool.
     * @param conn The connection.
     */
    public void doneWithConnection(HttpAuthConnection conn)
        throws HttpAuthException
    {
        synchronized(this)
        {
            if(m_checkedout > 0)
                m_checkedout--;
            
            if(conn.isConnected())
            {
                // TODO: I don't know of any way to drain the remaining
                // junk out of a socket here without blocking. :(
                
                m_connections.push(conn);
            }
        }
    }
}
