/*
 * Copyright (c) 2004, Stefan Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 *     * Redistributions of source code must retain the above 
 *       copyright notice, this list of conditions and the 
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the 
 *       above copyright notice, this list of conditions and 
 *       the following disclaimer in the documentation and/or 
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be 
 *       used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGE.
 * 
 *
 * CONTRIBUTORS
 *  Stef Walter <stef@memberwebs.com>
 *
 */ 
 
package com.memberwebs.httpauth.jetty;

import java.io.IOException;
import java.util.Enumeration;

import org.mortbay.http.HttpRequest;
import org.mortbay.http.HttpResponse;
import org.mortbay.http.UserPrincipal;
import org.mortbay.http.UserRealm;
import org.mortbay.http.SecurityConstraint.Authenticator;

import com.memberwebs.httpauth.BaseHttpAuthenticator;
import com.memberwebs.httpauth.HttpAuthConnectionSource;
import com.memberwebs.httpauth.HttpAuthException;
import com.memberwebs.httpauth.HttpAuthFilter;

/**
 * An authenticator for the Jetty web server. An instance of this object 
 * is used by a Jetty HttpContext to perform authentication.
 *
 * @author Stef Walter
 */
public class JettyHttpAuthenticator 
    extends BaseHttpAuthenticator 
    implements Authenticator
{
    protected HttpAuthFilter m_filter;
    
    /**
     * Creates a new JettyHttpAuthenticator. 
     * 
     * @param source A source of HttpAuth connections
     *               (such as {@link #HttpAuthConnectionPool})
     * @param filter An optional filter which determines 
     *               which types of authentication may be performed.
     */
    public JettyHttpAuthenticator(HttpAuthConnectionSource source, 
                                  HttpAuthFilter filter)
    {
        super(source);
        m_filter = filter;
    }
    
    /**
     * Called by Jetty's HttpContext to perform authentication.
     * @param realm
     * @param pathInContext
     * @param request
     * @param response
     * @return Used by Jetty to identify the user.
     */
    public UserPrincipal authenticated(UserRealm realm, String pathInContext, 
                                       HttpRequest request, HttpResponse response) 
        throws IOException
    {
        // We ignore the realm

        String[] authtypes = null;
        
        if(m_filter != null)
            authtypes = m_filter.getAuthTypes(request.getRemoteAddr(), request.isConfidential());

        String user = null;
        
        // Build a unique and consistent Connection ID so that NTLM works
        Object obj = request.getHttpConnection().getObject();
        if(obj == null)
        {
        	obj = "" + Math.random();
        	request.getHttpConnection().setObject(obj);
        }

        StringBuffer connid = new StringBuffer(32);
        connid.append(obj.toString());
        connid.append(":");
        connid.append(obj.hashCode());
        connid.append(":");
        connid.append(request.getHttpConnection().hashCode());

        try
        {
        	// Send off to httpauth for authentication
            user = authenticateRequest(request, response, connid.toString(), request.getMethod(), 
                                       request.getURI().toString(), authtypes);
        }
        catch(HttpAuthException e)
        {
            throw new IOException(e.getMessage());
        }
        
        if(user == null)
            return null;
        
        HttpAuthPrincipal principal = new HttpAuthPrincipal(user);
        
        request.setAuthType("HTTPAUTH");
        request.setAuthUser(user);
        request.setUserPrincipal(principal);
        
        return principal;
    }
    
    /**
     * Called by Jetty's HttpContext to get authentication method.
     * @return Used by Jetty to identify the method.
     */
    public String getAuthMethod()
    {
        return "HTTPAUTH";
    }
    
    /* -------------------------------------------------------------------------
     * OVERRIDES
     */

    protected void addResponseHeader(Object response, String name, String value)
    {
        HttpResponse resp = (HttpResponse)response;
        resp.addField(name, value);
    }

    protected void setResponseResult(Object response, int code)
        throws IOException
    {
        HttpResponse resp = (HttpResponse)response;

        if(code >= 300)
            resp.sendError(code);
        else
            resp.setStatus(code);
    }

    protected Enumeration getHeaderValues(Object request, String name)
    {
        HttpRequest req = (HttpRequest)request;
        return req.getFieldValues(name);
    }

    protected Enumeration getHeaderNames(Object request)
    {
        HttpRequest req = (HttpRequest)request;
        return req.getFieldNames();
    }
    
    protected class HttpAuthPrincipal 
        implements UserPrincipal
    {
        private String m_name;
    
        public HttpAuthPrincipal(String name)
        {
            m_name = name;
        }
        
        public boolean isAuthenticated()
        {
            return true;
        }

        public boolean isUserInRole(String role)    
        {
            return false;
        }

        public String getName()
        {
            return m_name;
        }
    }
    
    // A dumb ID to suppress compiler warnings
    private static final long serialVersionUID = -1266258707017895782L;
}
